#include <linux/module.h>
#include <linux/kprobes.h>
#include <linux/bio.h>
#include <linux/blkdev.h>
#include <linux/hash.h>
#include <linux/ide.h>
#include <linux/time.h>
#include <linux/fs.h>
#include <linux/syscalls.h>
#include <linux/kthread.h>
#include <linux/kernel.h>
#include <linux/file.h>
#include <asm/uaccess.h>

MODULE_AUTHOR("Sumit Narayan");
MODULE_LICENSE("GPL");

#define toggle(z) if(z==0){\
	z= 1;} else {z= 0;}

#define MAX_REC 5000
#define MAX_ENTRY 1
#define DEBUG_MODE

#ifdef DEBUG_MODE
#define DEBUGME printk
#else
#define DEBUGME(...)
#endif

typedef struct diskTrace
{
	sector_t		block;		/* Block requested		*/
	unsigned long	nr_sectors;	/* Number of sectors	*/
	struct timespec op_time;	/* Time of operation	*/
	unsigned int	op;			/* Read/Write			*/
} diskTrace;

int ctr = 0, count = 0;
int last_flag = 0;
int entryCtr = 1, fileCtr = 1;

char fname[10];

struct task_struct *ts;
struct diskTrace trace[2][25000];
struct file *fp = NULL;

static unsigned long no_of_rec = 0;

int trace_write(void *);

static void disk_trace(ide_drive_t *drive, struct request *rq, sector_t block)
{
    int flag = 0;
	DEBUGME("(dTrace): Count:%d Device %s: %sing: block=%llu, sectors=%lu, buffer=%d\n", 
            count, drive->name, rq_data_dir(rq) == READ ? "read" : "writ", block, 
            rq->nr_sectors, sizeof(rq->buffer));

	if(!strcmp(drive->name, "hda"))
	{	
		trace[flag][count].block = block;
		trace[flag][count].nr_sectors = rq->nr_sectors;
		trace[flag][count].op_time = CURRENT_TIME;
		trace[flag][count].op = (rq_data_dir(rq) == READ ? READ : WRITE);
		++count;
	}

    if(count >= MAX_REC)
    {
        count = 0;
        ts = kthread_run(trace_write, NULL, "dTrace_FileWrite");
    }

    ctr = flag;
    
    toggle(flag);
    jprobe_return();
}

int trace_write(void *ptr)
{
    int i = 0, a = 0;
    int target = MAX_REC;
	mm_segment_t fs;
    char buffer[50];

    if(fp == NULL)
    {
        sprintf(fname, "/root/trace%d.txt", fileCtr);
        fileCtr++;
        fp = filp_open(fname, O_CREAT | O_WRONLY, 0600);
    }
    
    fs = get_fs();
    set_fs(KERNEL_DS);

    /*
       Check if this is the exit time.
    */
    if(last_flag == 1)
    {
        target = count;
        last_flag = 0;
    }
    
    for(i=0;i<target;i++)
    {
        fp->f_pos = no_of_rec * 50; 
        sprintf(buffer, "%ld.%ld %llu %d %lu", trace[ctr][i].op_time.tv_sec, trace[ctr][i].op_time.tv_nsec,
                trace[ctr][i].block, trace[ctr][i].op, trace[ctr][i].nr_sectors);
        a = fp->f_op->write(fp, buffer, sizeof(buffer), &fp->f_pos);
        no_of_rec++;
    }
    set_fs(fs);

    /*
       To change to a new file, after MAXENTRY number of times,
       the file has been appended.
    */
    entryCtr++;
    if(entryCtr > MAX_ENTRY)
    {
        fp = NULL;
        entryCtr = 1;
    }

    return 0;
}

static struct jprobe my_jprobe = {
	.kp.addr = (kprobe_opcode_t *) 0xc03ae3b0, /* for __ide_do_rw_disk */
	.entry = (kprobe_opcode_t *) disk_trace
};

int init_module(void)
{
	register_jprobe(&my_jprobe);
	DEBUGME("(dTrace): dTrace planted. Please unload this module after completing.\n");
	
    fp = filp_open("/root/trace1.txt", O_CREAT | O_WRONLY | O_APPEND, 0600);
    if(!fp || !fp->f_op || !fp->f_op->write)
    {
        printk("(dTrace): WARNING -- Unable to open file for writing.\n");
        fileCtr = 0;
    }
    
    fp->f_pos = 0;
	return 0;
}

void cleanup_module(void)
{
	unregister_jprobe(&my_jprobe);
    last_flag = 1;
    /*
       For the remaining trace.
    */
    if(count>0)
    {
        DEBUGME("(dTrace): Writing remaining trace records.(%d)\n", count);
        trace_write(NULL);
    }
    filp_close(fp, NULL);
	DEBUGME("(dTrace): dTrace removed. Thank you for using dTrace.\n");
    DEBUGME("(dTrace): Please email sumitn@engr.uconn.edu for bugs.\n");
}

